# -*- coding: utf-8 -*-

import logging
import os

import click
import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from sklearn.ensemble import VotingClassifier
from sklearn.metrics import accuracy_score
from sklearn.model_selection import train_test_split, KFold
from sklearn.preprocessing import MinMaxScaler

from sklearn.utils import resample, shuffle

from Classifiers import DefinedClassifiers
from FeatureExpansion import Functions


def load_dataset(path, records_per_second):
    """
    Function to load sensor data as pandas data frame. The data file should be in .csv format, with 8 columns:
    <acc_x,acc_y,acc_z,gyro_x,gyro_y,gyro_z,timestamp,label>
    One can either input the path to a single file, or a directory containing a couple of csv files, which were
    collected in parallel from different sensors of a forklift.
    """

    print("Loading dataset from '{}'".format(path))

    dataframes = []

    dtype = {
        'ACC_X': np.float,
        'ACC_Y': np.float,
        'ACC_Z': np.float,
        'GYRO_X': np.float,
        'GYRO_Y': np.float,
        'GYRO_Z': np.float,
        'Realtime': str,
        'Labels': str
    }

    rename = ['ACC_X', 'ACC_Y', 'ACC_Z', 'GYRO_X', 'GYRO_Y', 'GYRO_Z']

    if os.path.isfile(path) and ".csv" in path:  # single csv file
        try:
            # df = pd.read_csv(path, dtype=dtype, parse_dates=parse_dates)
            df = pd.read_csv(path, dtype=dtype)
            dataframes.append(df)
        except IOError:
            logging.warning("Failed to read file: " + path)

    elif os.path.isdir(path):  # directory of csv files
        try:
            for fileName in os.listdir(path):
                file_path = os.path.join(path, fileName)
                if os.path.isfile(file_path) and ".csv" in file_path:
                    df = pd.read_csv(file_path, dtype=dtype)
                    dataframes.append(df)
        except IOError:
            logging.warning("Failed to read files from directory: " + path)
    else:
        print("Please input either the path of a csv file, or the directory path of a list of csv files.")

    target = pd.DataFrame()
    if len(dataframes) != 0:
        print("Joining {} datasets".format(len(dataframes)))
        interpolated = [interpolation(df, records_per_second) for df in dataframes]
        interpolated = [df.rename(columns={name: '{}_{}'.format(name, i) for name in rename}) for i, df in
                        enumerate(interpolated)]  # make column names unique
        target = pd.concat(interpolated, axis=1, join="outer").reset_index()

    print("Finished loading dataset")

    target = target.loc[:, ~target.columns.duplicated()]

    return target


def interpolation(df, records_per_second):
    """
    Process the dataset, to ensure that there are always the same amount of samples for each second, the number of
    records per second is the parameter. The current implementation divides the data set into records_per_second
    portions, taking the average value of each portion.
    """

    print("Interpolating dataset")

    seconds = df.Realtime.unique()
    target = []

    labels = pd.concat([df["Labels"], df["Realtime"]], axis=1)

    for sec in seconds:
        size = df[df.Realtime == sec].shape[0]
        if size >= records_per_second:
            unit = size // records_per_second
            for i in range(records_per_second):
                if i != records_per_second - 1:
                    tmp = df[df.Realtime == sec][unit * i:unit * (i + 1)].mean().astype("int")
                else:
                    tmp = df[df.Realtime == sec][unit * i:].mean().astype("int")
                tmp["Realtime"] = sec
                target.append(tmp)
        else:
            for _ in range(records_per_second):
                tmp = df[df.Realtime == sec].mean().astype("int")
                tmp["Realtime"] = sec
                target.append(tmp)

    target = pd.concat([pd.DataFrame(target).set_index("Realtime"), labels.set_index("Realtime")], axis=1, join="outer")
    return target


def expand_feature_space(df, window_length, overlap):
    """
    Given a dataframe, apply a sliding window with given parameters. For each sliding window, expand the feature space
    of the window with different statistical features. The label of the window is currently determined by the majority.
    """

    raw_list = []
    label_list = []
    mean_list = []
    std_list = []
    var_list = []
    skew_list = []
    kurt_list = []
    max_list = []
    med_list = []
    min_list = []

    dataset_windows = generate_windows(df, window_length, overlap)

    for window in dataset_windows:
        raw_list.append(window)
        label_list.append(max(set(window.Labels.values), key=list(window.Labels.values).count))
        mean_list.append(Functions.add_mean(window))
        std_list.append(Functions.add_std(window))
        var_list.append(Functions.add_variance(window))
        skew_list.append(Functions.add_skew(window))
        kurt_list.append(Functions.add_kurtosis(window))
        max_list.append(Functions.add_max(window))
        med_list.append(Functions.add_median(window))
        min_list.append(Functions.add_min(window))

    targets = [pd.DataFrame(mean_list), pd.DataFrame(std_list), pd.DataFrame(var_list), pd.DataFrame(skew_list),
               pd.DataFrame(kurt_list), pd.DataFrame(max_list), pd.DataFrame(med_list), pd.DataFrame(min_list),
               pd.DataFrame(label_list)]

    # todo create column names
    return pd.concat(targets, axis=1).round(3), pd.DataFrame(raw_list), raw_list


def generate_windows(df, window_length, overlap):
    """
    Given a dataframe, the parameters window length and overlap rate, output a list of windows.
    Window shape (winLen, df.shape[1])
    """

    step = max(int(window_length * (1 - overlap)), 1)

    windows = []
    for i in range(0, len(df) - window_length, step):
        windows.append(df[i:i + window_length].reset_index(drop=True))
    return windows


def transorm_dtw_data(df):
    X_ = [d[d.columns.difference(["Labels", "Realtime"])].values.T for d in df]
    X__ = np.array(X_)

    shape = X__.shape[:-2] + (-1,)

    X = np.reshape(X__, shape)
    y_ = [d.Labels[0] for d in df]
    y = np.array(y_)

    return X, y


def generate_cross_validation_sets(df, fold=10):
    """
    Generate a list of train- and testset pairs according to the K-fold Cross Validation
    The given dataframe contains N-1 column features and the last column as the label
    """
    target = []

    if isinstance(df, pd.DataFrame):
        X = df.iloc[:, :-1].values
        y = df.iloc[:, -1].values
    else:
        X, y = transorm_dtw_data(df)

    kf = KFold(n_splits=fold, random_state=None, shuffle=False)
    for train_index, valid_index in kf.split(X):
        X_train, X_valid = X[train_index], X[valid_index]
        y_train, y_valid = y[train_index], y[valid_index]
        target.append([X_train, X_valid, y_train, y_valid])
    return target


def upsampling(df):
    # resample the dataset to a balance distributed form, so that every class has the same amount of instance as the
    # majority class
    label_count = {}
    target_buf = []
    for label in df.iloc[:, -1].unique():
        label_count[label] = df.iloc[:, -1].value_counts()[label]
    majority_amount = max(label_count.values())
    majority_class = max(label_count, key=label_count.get)
    for minorityClass in label_count.keys():
        if minorityClass is not majority_class:
            target_buf.append(resample(df[df.iloc[:, -1] == minorityClass], n_samples=majority_amount))
        else:
            target_buf.append(df[df.iloc[:, -1] == minorityClass])
    return shuffle(pd.concat(target_buf))


def normalization(df_list):
    # all given dataframes will be normalized to range (-1,1) with the same scaler

    scaler = MinMaxScaler(feature_range=(-1, 1))
    data = df_list[0].iloc[:, :-1]
    scaler.fit(data)

    target = []
    for df in df_list:
        tmp = pd.concat([pd.DataFrame(scaler.transform(df.iloc[:, :-1])).reset_index(drop=True),
                         df.iloc[:, -1].reset_index(drop=True)], axis=1)
        target.append(tmp)
    return target


@click.command()
@click.argument("input-path", type=click.Path())
@click.option("--records-per-second", default=3, help="Number of records per second after resampling")
@click.option("--window-length", default=10, help="Size of the sliding window")
@click.option("--overlap", default=0.5, help="Relative overlap of the sliding windows")
@click.option("--test-size", default=0.2, help="Relative size of the test split")
@click.option("--seed", default=42, help="Random seed")
@click.option("--n-components", default=10, help="Number of components in the PCA")
@click.option("--k-fold", default=5, help="Number of iterations for cross validation")
@click.option("--voting", default=True, help="Create ensemble classifier")
@click.option("--feature", default=True, help="Test feature classifiers")
@click.option("--dtw", default=True, help="Test DTW classifiers")
def main(**kwargs):
    path = kwargs['input_path']
    records_per_second = kwargs['records_per_second']
    window_length = kwargs['window_length']
    overlap = kwargs['overlap']
    test_size = kwargs['test_size']
    seed = kwargs['seed']
    n_components = kwargs['n_components']
    k_fold = kwargs['k_fold']
    voting = kwargs['voting']
    feature = kwargs['feature']
    dtw = kwargs['dtw']

    df = load_dataset(path, records_per_second)
    df1, df1_dtw, df1_dtw_list = expand_feature_space(df, window_length, overlap)

    # Split windows into training- and testsets
    trainset_raw, testset_raw, trainset_dtw_raw, testset_dtw_raw = train_test_split(df1, df1_dtw_list, test_size=test_size,
                                                                                    random_state=seed)

    # scaling the dataset to [-1,1]
    trainset, testset = normalization([trainset_raw, testset_raw])
    trainset_dtw, testset_dtw = trainset_dtw_raw, testset_dtw_raw

    # resample the biased trainset
    trainset = upsampling(trainset)

    # Apply PCA
    trainset = trainset.reset_index(drop=True)
    testset = testset.reset_index(drop=True)

    pca = PCA(n_components=n_components)
    pca.fit(trainset.iloc[:, :-1])
    train_data = pd.DataFrame(pca.transform(trainset.iloc[:, :-1]))
    test_data = pd.DataFrame(pca.transform(testset.iloc[:, :-1]))

    trainset = pd.concat([train_data, trainset.iloc[:, -1]], axis=1)
    testset = pd.concat([test_data, testset.iloc[:, -1]], axis=1)

    # Apply cross validation, get the dataset of each fold
    cross_validation_set = generate_cross_validation_sets(trainset.reset_index(drop=True), fold=k_fold)
    cross_validation_set_dtw = generate_cross_validation_sets(trainset_dtw, fold=k_fold)

    # iterate classical models, training and validate over each fold
    result = []
    ensemble_classifiers = []

    if feature:
        for i, (clfName, classifier) in enumerate(DefinedClassifiers.classifiers.items()):
            print("Statistical feature classifier {}/{}".format(i + 1, len(DefinedClassifiers.classifiers)))

            clf = classifier()
            ensemble_classifiers.append((clfName, clf))
            clf, score = DefinedClassifiers.cv_classifier(clf, cross_validation_set)

            print("Average validation accuracy of {} is: {}".format(clfName, score))

            predicted_labels = clf.predict(testset.iloc[:, :-1].values)
            accuracy = accuracy_score(testset.iloc[:, -1].values, predicted_labels)

            print("Test accuracy of {} is: {}".format(clfName, accuracy))

            result.append([clfName, accuracy])

    if dtw:
        for i, (clfName, classifier) in enumerate(DefinedClassifiers.knnDtwClassifiers.items()):
            print("DTW classifier {}/{}".format(i + 1, len(DefinedClassifiers.knnDtwClassifiers)))

            clf = classifier()
            clf.metric.num_features = window_length
            clf, score = DefinedClassifiers.cv_classifier(clf, cross_validation_set_dtw)

            print("Average validation accuracy of {} is: {}".format(clfName, score))

            X_test, y_test = transorm_dtw_data(testset_dtw)
            predicted_labels = clf.predict(X_test)
            accuracy = accuracy_score(y_test, predicted_labels)

            print("Test accuracy of {} is: {}".format(clfName, accuracy))

            result.append([clfName, accuracy])

    if voting:
        if not feature:
            print("Skipping voting classifier, because feature classifieres were disabled")
        else:
            print("Voting classifier from {} feature classifiers".format(len(ensemble_classifiers)))
            eclf = VotingClassifier(estimators=ensemble_classifiers, voting='hard')
            eclf, escore = DefinedClassifiers.cv_classifier(eclf, cross_validation_set)
            print("Average validation accuracy of ensembeled classifier is: ", escore)

            epredicted_labels = eclf.predict(testset.iloc[:, :-1].values)
            eaccuracy = accuracy_score(testset.iloc[:, -1].values, epredicted_labels)
            print("Test accuracy of ensembeled classifier is: ", eaccuracy)


if __name__ == "__main__":
    main()
